/*
 * livepatch_bsc1205186
 *
 * Fix for CVE-2022-2602, bsc#1205186
 *
 *  Upstream commit:
 *  0091bfc81741 ("io_uring/af_unix: defer registered files gc to io_uring
 *                 release")
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 and -SP3 commit:
 *  fee586221f0512c91a9b9ab0315a6fdb83cae687
 *
 *  SLE15-SP4 and SLE15-SP4-RT commit:
 *  52299ff5af1b66700865c14aabbc365c5de9beb5
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from fs/io_uring.c */
static const struct file_operations (*klpe_io_uring_fops);

/* klp-ccp: from net/unix/garbage.c */
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/un.h>
#include <linux/net.h>
#include <linux/fs.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/file.h>
#include <linux/mutex.h>
#include <linux/wait.h>
#include <net/sock.h>
#include <net/af_unix.h>

/* klp-ccp: from net/unix/garbage.c */
#include <net/scm.h>
#include <net/tcp_states.h>
#include <net/tcp_states.h>

/* klp-ccp: from net/unix/scm.h */
extern struct list_head gc_inflight_list;
extern spinlock_t unix_gc_lock;

/* klp-ccp: from net/unix/garbage.c */
static struct list_head (*klpe_gc_candidates);
static struct wait_queue_head (*klpe_unix_gc_wait);

static void (*klpe_scan_children)(struct sock *x, void (*func)(struct unix_sock *),
			  struct sk_buff_head *hitlist);

static void (*klpe_dec_inflight)(struct unix_sock *usk);

static void (*klpe_inc_inflight)(struct unix_sock *usk);

static void (*klpe_inc_inflight_move_tail)(struct unix_sock *u);

static bool (*klpe_gc_in_progress);

/* New. */
static bool klpp_skb_scm_io_uring(const struct sk_buff *skb)
{
	const struct sock *sk = skb->sk;

	if (!sk)
		return false;
	else if (!sk->sk_socket)
		return false;
	else if (!sk->sk_socket->file)
		return false;

	return sk->sk_socket->file->f_op == &(*klpe_io_uring_fops);
}

void klpp_unix_gc(void)
{
	/*
	 * Fix CVE-2022-2602
	 *  +1 line
	 */
	struct sk_buff *next_skb, *skb;
	struct unix_sock *u;
	struct unix_sock *next;
	struct sk_buff_head hitlist;
	struct list_head cursor;
	LIST_HEAD(not_cycle_list);

	spin_lock(&unix_gc_lock);

	/* Avoid a recursive GC. */
	if ((*klpe_gc_in_progress))
		goto out;

	(*klpe_gc_in_progress) = true;
	/* First, select candidates for garbage collection.  Only
	 * in-flight sockets are considered, and from those only ones
	 * which don't have any external reference.
	 *
	 * Holding unix_gc_lock will protect these candidates from
	 * being detached, and hence from gaining an external
	 * reference.  Since there are no possible receivers, all
	 * buffers currently on the candidates' queues stay there
	 * during the garbage collection.
	 *
	 * We also know that no new candidate can be added onto the
	 * receive queues.  Other, non candidate sockets _can_ be
	 * added to queue, so we must make sure only to touch
	 * candidates.
	 */
	list_for_each_entry_safe(u, next, &gc_inflight_list, link) {
		long total_refs;
		long inflight_refs;

		total_refs = file_count(u->sk.sk_socket->file);
		inflight_refs = atomic_long_read(&u->inflight);

		BUG_ON(inflight_refs < 1);
		BUG_ON(total_refs < inflight_refs);
		if (total_refs == inflight_refs) {
			list_move_tail(&u->link, &(*klpe_gc_candidates));
			__set_bit(UNIX_GC_CANDIDATE, &u->gc_flags);
			__set_bit(UNIX_GC_MAYBE_CYCLE, &u->gc_flags);
		}
	}

	/* Now remove all internal in-flight reference to children of
	 * the candidates.
	 */
	list_for_each_entry(u, &(*klpe_gc_candidates), link)
		(*klpe_scan_children)(&u->sk, (*klpe_dec_inflight), NULL);

	/* Restore the references for children of all candidates,
	 * which have remaining references.  Do this recursively, so
	 * only those remain, which form cyclic references.
	 *
	 * Use a "cursor" link, to make the list traversal safe, even
	 * though elements might be moved about.
	 */
	list_add(&cursor, &(*klpe_gc_candidates));
	while (cursor.next != &(*klpe_gc_candidates)) {
		u = list_entry(cursor.next, struct unix_sock, link);

		/* Move cursor to after the current position. */
		list_move(&cursor, &u->link);

		if (atomic_long_read(&u->inflight) > 0) {
			list_move_tail(&u->link, &not_cycle_list);
			__clear_bit(UNIX_GC_MAYBE_CYCLE, &u->gc_flags);
			(*klpe_scan_children)(&u->sk, (*klpe_inc_inflight_move_tail), NULL);
		}
	}
	list_del(&cursor);

	/* Now gc_candidates contains only garbage.  Restore original
	 * inflight counters for these as well, and remove the skbuffs
	 * which are creating the cycle(s).
	 */
	skb_queue_head_init(&hitlist);
	list_for_each_entry(u, &(*klpe_gc_candidates), link)
		(*klpe_scan_children)(&u->sk, (*klpe_inc_inflight), &hitlist);

	/* not_cycle_list contains those sockets which do not make up a
	 * cycle.  Restore these to the inflight list.
	 */
	while (!list_empty(&not_cycle_list)) {
		u = list_entry(not_cycle_list.next, struct unix_sock, link);
		__clear_bit(UNIX_GC_CANDIDATE, &u->gc_flags);
		list_move_tail(&u->link, &gc_inflight_list);
	}

	spin_unlock(&unix_gc_lock);

	/*
	 * Fix CVE-2022-2602
	 *  +12 lines
	 */
	/* We need io_uring to clean its registered files, ignore all io_uring
	 * originated skbs. It's fine as io_uring doesn't keep references to
	 * other io_uring instances and so killing all other files in the cycle
	 * will put all io_uring references forcing it to go through normal
	 * release.path eventually putting registered files.
	 */
	skb_queue_walk_safe(&hitlist, skb, next_skb) {
		if (klpp_skb_scm_io_uring(skb)) {
			__skb_unlink(skb, &hitlist);
			skb_queue_tail(&skb->sk->sk_receive_queue, skb);
		}
	}

	/* Here we are. Hitlist is filled. Die. */
	__skb_queue_purge(&hitlist);

	spin_lock(&unix_gc_lock);

	/*
	 * Fix CVE-2022-2602
	 *  +5 lines
	 */
	 /* There could be io_uring registered files, just push them back to
	 * the inflight list
	 */
	list_for_each_entry_safe(u, next, &(*klpe_gc_candidates), link)
		list_move_tail(&u->link, &gc_inflight_list);


	/* All candidates should have been detached by now. */
	BUG_ON(!list_empty(&(*klpe_gc_candidates)));
	(*klpe_gc_in_progress) = false;
	wake_up(&(*klpe_unix_gc_wait));

 out:
	spin_unlock(&unix_gc_lock);
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1205186.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "dec_inflight", (void *)&klpe_dec_inflight },
	{ "gc_candidates", (void *)&klpe_gc_candidates },
	{ "gc_in_progress", (void *)&klpe_gc_in_progress },
	{ "inc_inflight", (void *)&klpe_inc_inflight },
	{ "inc_inflight_move_tail", (void *)&klpe_inc_inflight_move_tail },
	{ "io_uring_fops", (void *)&klpe_io_uring_fops },
	{ "scan_children", (void *)&klpe_scan_children },
	{ "unix_gc_wait", (void *)&klpe_unix_gc_wait },
};

int livepatch_bsc1205186_init(void)
{
	return __klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
